@@ -59,9 +59,11 @@ struct _BluetoothClient {
 
 	GListStore *list_store;
 	Adapter1 *default_adapter;
+	gboolean has_power_state;
 	GDBusObjectManager *manager;
 	GCancellable *cancellable;
 	guint num_adapters;
+	/* Prevent concurrent enabling or disabling discoveries */
 	gboolean discovery_started;
 	UpClient *up_client;
 	gboolean bluez_devices_coldplugged;
@@ -74,6 +76,7 @@ enum {
 	PROP_NUM_ADAPTERS,
 	PROP_DEFAULT_ADAPTER,
 	PROP_DEFAULT_ADAPTER_POWERED,
+	PROP_DEFAULT_ADAPTER_STATE,
 	PROP_DEFAULT_ADAPTER_SETUP_MODE,
 	PROP_DEFAULT_ADAPTER_NAME,
 	PROP_DEFAULT_ADAPTER_ADDRESS
@@ -111,6 +114,26 @@ get_device_for_path (BluetoothClient *client,
 	return NULL;
 }
 
+static BluetoothDevice *
+get_device_for_bdaddr (BluetoothClient *client,
+		       const char      *bdaddr)
+{
+	guint n_items, i;
+
+	n_items = g_list_model_get_n_items (G_LIST_MODEL (client->list_store));
+	for (i = 0; i < n_items; i++) {
+		g_autoptr(BluetoothDevice) d = NULL;
+		g_autofree char *s = NULL;
+
+		d = g_list_model_get_item (G_LIST_MODEL (client->list_store), i);
+		g_object_get (G_OBJECT (d), "address", &s, NULL);
+		if (g_ascii_strncasecmp (bdaddr, s, BDADDR_STR_LEN) == 0) {
+			return g_steal_pointer (&d);
+		}
+	}
+	return NULL;
+}
+
 static char **
 device_list_uuids (const gchar * const *uuids)
 {
@@ -553,6 +576,10 @@ adapter_notify_cb (Adapter1       *adapter,
 		g_object_notify (G_OBJECT (client), "default-adapter-setup-mode");
 	} else if (g_strcmp0 (property, "powered") == 0) {
 		g_object_notify (G_OBJECT (client), "default-adapter-powered");
+		if (!client->has_power_state)
+			g_object_notify (G_OBJECT (client), "default-adapter-state");
+	} else if (g_strcmp0 (property, "power-state") == 0) {
+		g_object_notify (G_OBJECT (client), "default-adapter-state");
 	}
 }
 
@@ -569,6 +596,7 @@ notify_default_adapter_props (BluetoothClient *client)
 	g_object_notify (G_OBJECT (client), "default-adapter");
 	g_object_notify (G_OBJECT (client), "default-adapter-address");
 	g_object_notify (G_OBJECT (client), "default-adapter-powered");
+	g_object_notify (G_OBJECT (client), "default-adapter-state");
 	g_object_notify (G_OBJECT (client), "default-adapter-setup-mode");
 	g_object_notify (G_OBJECT (client), "default-adapter-name");
 }
@@ -703,7 +731,6 @@ adapter_removed (GDBusObjectManager   *manager,
 	if (!was_default)
 		goto out;
 
-
 	new_default_adapter = NULL;
 	object_list = g_dbus_object_manager_get_objects (client->manager);
 	for (l = object_list; l != NULL; l = l->next) {
@@ -976,7 +1003,7 @@ up_device_added_cb (UpClient *up_client,
 		    gpointer  user_data)
 {
 	BluetoothClient *client = user_data;
-	g_autofree char *native_path = NULL;
+	g_autofree char *serial = NULL;
 	g_autoptr(BluetoothDevice) device = NULL;
 	UpDeviceLevel battery_level;
 	double percentage;
@@ -985,16 +1012,16 @@ up_device_added_cb (UpClient *up_client,
 	g_debug ("Considering UPower device %s", up_device_get_object_path (up_device));
 
 	g_object_get (up_device,
-		      "native-path", &native_path,
+		      "serial", &serial,
 		      "battery-level", &battery_level,
 		      "percentage", &percentage,
 		      NULL);
 
-	if (!native_path || !g_str_has_prefix (native_path, "/org/bluez/"))
+	if (!serial || !bluetooth_verify_address(serial))
 		return;
-	device = get_device_for_path (client, native_path);
+	device = get_device_for_bdaddr (client, serial);
 	if (!device) {
-		g_debug ("Could not find bluez device for upower device %s", native_path);
+		g_debug ("Could not find bluez device for upower device with serial %s", serial);
 		return;
 	}
 	g_signal_connect (G_OBJECT (up_device), "notify::battery-level",
@@ -1006,7 +1033,7 @@ up_device_added_cb (UpClient *up_client,
 		battery_type = BLUETOOTH_BATTERY_TYPE_PERCENTAGE;
 	else
 		battery_type = BLUETOOTH_BATTERY_TYPE_COARSE;
-	g_debug ("Applying battery information for %s", native_path);
+	g_debug ("Applying battery information for %s", serial);
 	g_object_set (device,
 		      "battery-type", battery_type,
 		      "battery-level", battery_level,
@@ -1078,6 +1105,7 @@ static void bluetooth_client_init(BluetoothClient *client)
 {
 	client->cancellable = g_cancellable_new ();
 	client->list_store = g_list_store_new (BLUETOOTH_TYPE_DEVICE);
+	client->has_power_state = TRUE;
 
 	g_dbus_object_manager_client_new_for_bus (G_BUS_TYPE_SYSTEM,
 						  G_DBUS_OBJECT_MANAGER_CLIENT_FLAGS_DO_NOT_AUTO_START,
@@ -1103,6 +1131,45 @@ _bluetooth_client_get_default_adapter(BluetoothClient *client)
 	return G_DBUS_PROXY (g_object_ref (client->default_adapter));
 }
 
+static void
+stop_discovery_cb (Adapter1     *adapter,
+		   GAsyncResult *res,
+		   gpointer      user_data)
+{
+	g_autoptr(GError) error = NULL;
+	gboolean ret;
+
+	ret = adapter1_call_stop_discovery_finish (adapter, res, &error);
+	if (!ret) {
+		g_debug ("Error calling StopDiscovery() on %s org.bluez.Adapter1: %s (%s, %d)",
+			 g_dbus_proxy_get_object_path (G_DBUS_PROXY (adapter)),
+			 error->message, g_quark_to_string (error->domain), error->code);
+	} else {
+		g_debug ("Ran StopDiscovery() successfully on %s org.bluez.Adapter1",
+			 g_dbus_proxy_get_object_path (G_DBUS_PROXY (adapter)));
+	}
+}
+
+static void
+start_discovery_cb (Adapter1     *adapter,
+		    GAsyncResult *res,
+		    gpointer      user_data)
+{
+	g_autoptr(GError) error = NULL;
+	gboolean ret;
+
+	ret = adapter1_call_start_discovery_finish (adapter, res, &error);
+	if (!ret) {
+		g_debug ("Error calling StartDiscovery() on %s org.bluez.Adapter1: %s (%s, %d)",
+			 g_dbus_proxy_get_object_path (G_DBUS_PROXY (adapter)),
+			 error->message, g_quark_to_string (error->domain), error->code);
+		if (!g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED)) {
+			BluetoothClient *client = user_data;
+			client->discovery_started = FALSE;
+		}
+	}
+}
+
 static void
 set_discovery_filter_cb (Adapter1     *adapter,
 			 GAsyncResult *res,
@@ -1113,33 +1180,20 @@ set_discovery_filter_cb (Adapter1     *adapter,
 
 	ret = adapter1_call_set_discovery_filter_finish (adapter, res, &error);
 	if (!ret) {
-		g_debug ("Error calling SetDiscoveryFilter()  on interface org.bluez.Adapter1: %s (%s, %d)",
-			 error->message, g_quark_to_string (error->domain), error->code);
-	}
-}
-
-static void
-discovery_cb (Adapter1     *adapter,
-	      GAsyncResult *res,
-	      gpointer      user_data)
-{
-	g_autoptr(GError) error = NULL;
-	gboolean ret;
-	gboolean start_call = GPOINTER_TO_UINT (user_data);
-
-	if (start_call)
-		ret = adapter1_call_start_discovery_finish (adapter, res, &error);
-	else
-		ret = adapter1_call_stop_discovery_finish (adapter, res, &error);
-	if (!ret) {
-		g_debug ("Error calling %s on %s org.bluez.Adapter1: %s (%s, %d)",
-			 start_call ? "StartDiscovery()" : "StopDiscovery()",
-			 g_dbus_proxy_get_object_path (G_DBUS_PROXY (adapter)),
+		g_debug ("Error calling SetDiscoveryFilter() on interface org.bluez.Adapter1: %s (%s, %d)",
 			 error->message, g_quark_to_string (error->domain), error->code);
+		if (!g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED)) {
+			BluetoothClient *client = user_data;
+			client->discovery_started = FALSE;
+		}
 	} else {
-		g_debug ("Ran %s successfully on %s org.bluez.Adapter1",
-			 start_call ? "StartDiscovery()" : "StopDiscovery()",
-			 g_dbus_proxy_get_object_path (G_DBUS_PROXY (adapter)));
+		BluetoothClient *client = user_data;
+
+		g_debug ("Starting discovery on %s", g_dbus_proxy_get_object_path (G_DBUS_PROXY (adapter)));
+		adapter1_call_start_discovery (ADAPTER1 (adapter),
+					       client->cancellable,
+					       (GAsyncReadyCallback) start_discovery_cb,
+					       client);
 	}
 }
 
@@ -1157,6 +1211,7 @@ _bluetooth_client_set_default_adapter_discovering (BluetoothClient *client,
 	if (adapter == NULL) {
 		g_debug ("%s discovery requested, but no default adapter",
 			 discovering ? "Starting" : "Stopping");
+		client->discovery_started = FALSE;
 		return;
 	}
 
@@ -1171,25 +1226,47 @@ _bluetooth_client_set_default_adapter_discovering (BluetoothClient *client,
 						    client->cancellable,
 						    (GAsyncReadyCallback) set_discovery_filter_cb,
 						    client);
-	}
-
-	if (discovering) {
-		g_debug ("Starting discovery on %s", g_dbus_proxy_get_object_path (adapter));
-		adapter1_call_start_discovery (ADAPTER1 (adapter),
-					       client->cancellable,
-					       (GAsyncReadyCallback) discovery_cb,
-					       GUINT_TO_POINTER (discovering));
 	} else {
 		/* Don't cancel a discovery stop when the BluetoothClient
 		 * is finalised, so don't pass a cancellable */
 		g_debug ("Stopping discovery on %s", g_dbus_proxy_get_object_path (adapter));
 		adapter1_call_stop_discovery (ADAPTER1 (adapter),
 					      NULL,
-					      (GAsyncReadyCallback) discovery_cb,
-					      GUINT_TO_POINTER (discovering));
+					      (GAsyncReadyCallback) stop_discovery_cb,
+					      client);
 	}
 }
 
+static BluetoothAdapterState
+adapter_get_state (BluetoothClient *client)
+{
+	const char *str;
+
+	if (!client->default_adapter)
+		return BLUETOOTH_ADAPTER_STATE_ABSENT;
+
+	str = adapter1_get_power_state (client->default_adapter);
+	if (str) {
+		if (g_str_equal (str, "on"))
+			return BLUETOOTH_ADAPTER_STATE_ON;
+		if (g_str_equal (str, "off") ||
+		    g_str_equal (str, "off-blocked"))
+			return BLUETOOTH_ADAPTER_STATE_OFF;
+		if (g_str_equal (str, "off-enabling"))
+			return BLUETOOTH_ADAPTER_STATE_TURNING_ON;
+		if (g_str_equal (str, "on-disabling"))
+			return BLUETOOTH_ADAPTER_STATE_TURNING_OFF;
+		g_warning_once ("Unexpected adapter PowerState value '%s'", str);
+	} else {
+		client->has_power_state = FALSE;
+	}
+
+	/* Fallback if property is missing, or value is unexpected */
+	return adapter1_get_powered (client->default_adapter) ?
+		BLUETOOTH_ADAPTER_STATE_ON : BLUETOOTH_ADAPTER_STATE_OFF;
+
+}
+
 static void
 bluetooth_client_get_property (GObject        *object,
 			       guint           property_id,
@@ -1210,6 +1287,9 @@ bluetooth_client_get_property (GObject        *object,
 		g_value_set_boolean (value, client->default_adapter ?
 				     adapter1_get_powered (client->default_adapter) : FALSE);
 		break;
+	case PROP_DEFAULT_ADAPTER_STATE:
+		g_value_set_enum (value, adapter_get_state (client));
+		break;
 	case PROP_DEFAULT_ADAPTER_NAME:
 		g_value_set_string (value, client->default_adapter ?
 				    adapter1_get_alias (client->default_adapter) : NULL);
@@ -1343,6 +1423,18 @@ static void bluetooth_client_class_init(BluetoothClientClass *klass)
 					 g_param_spec_boolean ("default-adapter-powered", NULL,
 							      "Whether the default adapter is powered",
 							       FALSE, G_PARAM_READWRITE));
+	/**
+	 * BluetoothClient:default-adapter-state:
+	 *
+	 * The #BluetoothAdapterState of the default Bluetooth adapter. More precise than
+	 * #BluetoothClient:default-adapter-powered.
+	 */
+	g_object_class_install_property (object_class, PROP_DEFAULT_ADAPTER_STATE,
+					 g_param_spec_enum ("default-adapter-state", NULL,
+							    "State of the default adapter",
+							    BLUETOOTH_TYPE_ADAPTER_STATE,
+							    BLUETOOTH_ADAPTER_STATE_ABSENT,
+							    G_PARAM_READABLE));
 	/**
 	 * BluetoothClient:default-adapter-setup-mode:
 	 *
@@ -1676,6 +1768,14 @@ disconnect_callback (GDBusProxy   *proxy,
  * @callback: (scope async): a #GAsyncReadyCallback to call when the connection is complete
  * @user_data: the data to pass to callback function
  *
+ * This will start the process of connecting to one of the known-connectable
+ * services on the device. This means that it could connect to all the audio
+ * services on a headset, but just to the input service on a keyboard.
+ *
+ * Broadly speaking, this will only have an effect on devices with audio and HID
+ * services, and do nothing if the device doesn't have the "connectable"
+ * property set.
+ *
  * When the connection operation is finished, @callback will be called. You can
  * then call bluetooth_client_connect_service_finish() to get the result of the
  * operation.
