@@ -29,12 +29,13 @@
 #include <glib/gstdio.h>
 #include <glib/gi18n-lib.h>
 #include <gio/gio.h>
+#include <gsound.h>
 #include <gtk/gtk.h>
 #include <bluetooth-client.h>
 #include <libnotify/notify.h>
-#include <canberra-gtk.h>
 
 #include "bluetooth-settings-obexpush.h"
+#include "bluetooth-device.h"
 
 #define MANAGER_SERVICE	"org.bluez.obex"
 #define MANAGER_IFACE	"org.bluez.obex.AgentManager1"
@@ -67,6 +68,7 @@ G_DEFINE_TYPE(ObexAgent, obex_agent, G_TYPE_OBJECT)
 static ObexAgent *agent;
 static BluetoothClient *client;
 static GCancellable *cancellable;
+static GSoundContext *gsound_context;
 
 static void
 on_close_notification (NotifyNotification *notification)
@@ -130,7 +132,6 @@ show_notification (const char *filename)
 {
 	char *file_uri, *notification_text, *display, *mime_type;
 	NotifyNotification *notification;
-	ca_context *ctx;
 	GAppInfo *app;
 
 	file_uri = g_filename_to_uri (filename, NULL, NULL);
@@ -172,11 +173,12 @@ show_notification (const char *filename)
 	g_free (notification_text);
 
 	/* Now we do the audio notification */
-	ctx = ca_gtk_context_get ();
-	ca_context_play (ctx, 0,
-			 CA_PROP_EVENT_ID, "complete-download",
-			 CA_PROP_EVENT_DESCRIPTION, _("File reception complete"),
-			 NULL);
+	if (gsound_context) {
+		gsound_context_play_simple (gsound_context, cancellable, NULL,
+					    GSOUND_ATTR_EVENT_ID, "complete-download",
+					    GSOUND_ATTR_EVENT_DESCRIPTION, _("File reception complete"),
+					    NULL);
+	}
 }
 
 static void
@@ -269,51 +271,43 @@ get_paired_for_address (const char  *adapter,
 			const char  *device,
 			char       **name)
 {
-	GtkTreeModel *model;
-	GtkTreeIter parent;
-	gboolean next;
-	gboolean ret = FALSE;
-	char *addr;
+	g_autoptr(GListStore) model = NULL;
+	g_autofree char *default_address = NULL;
+	guint n_devices, i;
 
-	model = bluetooth_client_get_model (client);
-
-	for (next = gtk_tree_model_get_iter_first (model, &parent);
-	     next;
-	     next = gtk_tree_model_iter_next (model, &parent)) {
-		gtk_tree_model_get (model, &parent,
-			BLUETOOTH_COLUMN_ADDRESS, &addr,
-			-1);
-
-		if (g_strcmp0 (addr, adapter) == 0) {
-			GtkTreeIter child;
-			char *dev_addr;
-
-			for (next = gtk_tree_model_iter_children (model, &child, &parent);
-			     next;
-			     next = gtk_tree_model_iter_next (model, &child)) {
-				gboolean paired;
-				char *alias;
-
-				gtk_tree_model_get (model, &child,
-					BLUETOOTH_COLUMN_ADDRESS, &dev_addr,
-					BLUETOOTH_COLUMN_PAIRED, &paired,
-					BLUETOOTH_COLUMN_ALIAS, &alias,
-					-1);
-				if (g_strcmp0 (dev_addr, device) == 0) {
-					ret = paired;
-					*name = alias;
-					next = FALSE;
-				} else {
-					g_free (alias);
-				}
-				g_free (dev_addr);
-			}
-		}
-		g_free (addr);
+	g_object_get (client,
+		      "default-adapter-address", &default_address,
+		      NULL);
+	if (g_strcmp0 (default_address, adapter) != 0) {
+		g_debug ("Adapter '%s' is not the default adapter (%s)",
+			 adapter, default_address);
+		return FALSE;
 	}
 
-	g_object_unref (model);
-	return ret;
+	model = bluetooth_client_get_devices (client);
+	n_devices = g_list_model_get_n_items (G_LIST_MODEL (model));
+
+	for (i = 0; i < n_devices; i++) {
+		g_autoptr(BluetoothDevice) _device = NULL;
+		g_autofree char *address = NULL;
+		g_autofree char *alias = NULL;
+		gboolean paired = FALSE;
+
+		_device = g_list_model_get_item (G_LIST_MODEL (model), i);
+		g_object_get (_device,
+			      "address", &address,
+			      "alias", &alias,
+			      "paired", paired,
+			      NULL);
+		if (g_strcmp0 (device, address) != 0)
+			continue;
+
+		*name = g_steal_pointer (&alias);
+		return paired;
+	}
+
+	g_debug ("Could not find device '%s' in default adapter devices", device);
+	return FALSE;
 }
 
 static void
@@ -322,23 +316,21 @@ on_check_bonded_or_ask_session_acquired (GObject *object,
 					 gpointer user_data)
 {
 	GDBusMethodInvocation *invocation = user_data;
-	GDBusProxy *session;
-	GError *error = NULL;
+	g_autoptr(GDBusProxy) session = NULL;
+	g_autoptr(GError) error = NULL;
 	GVariant *v;
-	char *device, *adapter, *name;
+	g_autofree char *device = NULL;
+	g_autofree char *adapter = NULL;
+	g_autofree char *name = NULL;
 	gboolean paired;
 
 	session = g_dbus_proxy_new_for_bus_finish (res, &error);
 
 	if (!session) {
 		g_debug ("Failed to create a proxy for the session: %s", error->message);
-		g_clear_error (&error);
 		goto out;
 	}
 
-	device = NULL;
-	adapter = NULL;
-
 	/* obexd puts the remote device in Destination and our adapter
 	 * in Source */
 	v = g_dbus_proxy_get_cached_property (session, "Destination");
@@ -353,31 +345,22 @@ on_check_bonded_or_ask_session_acquired (GObject *object,
 		g_variant_unref (v);
 	}
 
-	g_object_unref (session);
-
 	if (!device || !adapter) {
 		g_debug ("Could not get remote device for the transfer");
-		g_free (device);
-		g_free (adapter);
 		goto out;
 	}
 
-	name = NULL;
 	paired = get_paired_for_address (adapter, device, &name);
-	g_free (device);
-	g_free (adapter);
 
 	if (paired) {
 		g_debug ("Remote device '%s' is paired, auto-accepting the transfer", name);
 		g_dbus_method_invocation_return_value (invocation,
 						       g_variant_new ("(s)", g_object_get_data (G_OBJECT (invocation), "temp-filename")));
-		g_free (name);
 		return;
 	} else {
 		ask_user (invocation,
 			  g_object_get_data (G_OBJECT (invocation), "filename"),
 			  name ? name : device);
-		g_free (name);
 		return;
 	}
 
@@ -852,4 +835,12 @@ obex_agent_up (void)
 
 	g_assert (cancellable == NULL);
 	cancellable = g_cancellable_new ();
+
+	if (gsound_context == NULL) {
+		g_autoptr(GError) error = NULL;
+
+		gsound_context = gsound_context_new (cancellable, &error);
+		if (!gsound_context)
+			g_warning ("Failed to open sound context: %s", error->message);
+	}
 }
